# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests the workspace UI helper."""

from django.test import RequestFactory, override_settings
from django.urls import reverse

from debusine.db.models import Workspace
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.ui.workspaces import WorkspaceUI


class WorkspaceUITests(TestCase):
    """Test the :py:class:`WorkspaceUI` class."""

    scenario = scenarios.DefaultContext(set_current=True)

    def get_helper(self, workspace: Workspace) -> WorkspaceUI:
        """Instantiate a UI helper for a workspace."""
        factory = RequestFactory()
        request = factory.get("/")
        return WorkspaceUI(request, workspace)

    def test_candidate_parents_empty(self) -> None:
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [])

    def test_candidate_parents_other_scope(self) -> None:
        other_scope = self.playground.get_or_create_scope("other")
        other_workspace = self.playground.create_workspace(
            name="other", scope=other_scope, public=True
        )
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [other_workspace])

    def test_candidate_parents_only_visible(self) -> None:
        self.playground.create_workspace(name="other")
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [])

    def test_candidate_parents_current_parents(self) -> None:
        """Current parents are candidates even if not normally accessible."""
        public = self.playground.create_workspace(name="public", public=True)
        private = self.playground.create_workspace(name="private")
        self.scenario.workspace.set_inheritance([private, public])
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [private, public])

    def test_candidate_parents_ordering(self) -> None:
        other_scope = self.playground.get_or_create_scope("other")
        b = self.playground.create_workspace(name="b", public=True)
        a = self.playground.create_workspace(name="a", public=True)
        c = self.playground.create_workspace(name="c", public=True)
        ob = self.playground.create_workspace(
            name="b", public=True, scope=other_scope
        )
        oa = self.playground.create_workspace(
            name="a", public=True, scope=other_scope
        )
        oc = self.playground.create_workspace(
            name="c", public=True, scope=other_scope
        )
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [a, b, c, oa, ob, oc])

    @override_settings(ROOT_URLCONF="debusine.web.archives.urls")
    def test_place_suite_list(self) -> None:
        helper = self.get_helper(self.scenario.workspace)
        place = helper.place_suite_list
        self.assertEqual(
            place.title, f"Suites in <code>{self.scenario.workspace}</code>"
        )
        self.assertEqual(
            place.url,
            reverse(
                "archive-root",
                kwargs={
                    "scope": self.scenario.scope.name,
                    "workspace": self.scenario.workspace.name,
                },
            ),
        )
        self.assertIs(place.parent, helper.place)
        self.assertEqual(place.breadcrumb, "suites")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for debusine templatetags."""

import datetime as dt
import re
from typing import ClassVar, cast, override
from unittest import mock

from django.contrib import messages
from django.contrib.auth.models import AnonymousUser
from django.contrib.messages.storage.base import Message
from django.db.models import Model
from django.template import TemplateSyntaxError, engines
from django.template.backends.django import Template as DjangoTemplate
from django.template.base import FilterExpression, Parser
from django.template.context import Context
from django.test import RequestFactory, override_settings
from django.utils.safestring import SafeString

from debusine.db.context import context
from debusine.db.models import Scope
from debusine.db.permissioncontext import PermissionContext, PermissionUser
from debusine.db.playground import scenarios
from debusine.server.middlewares.scopes import (
    TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL,
)
from debusine.test.django import TestCase
from debusine.web.helps import HELPS
from debusine.web.icons import Icons
from debusine.web.templatetags.debusine import (
    DebusineNode,
    DebusineNodeRenderError,
    PlaceNode,
    UINode,
    WidgetNode,
    _help,
    as_yaml,
    has_perm,
    icon,
    message_toast_color_class,
    message_toast_header_icon,
    roles,
    sorted_,
    worker_last_seen,
)
from debusine.web.views.base import Widget


class DebusineNodeTests(TestCase):
    """Tests for the :py:class:`DebusineNode` class."""

    def make_node(self) -> DebusineNode:
        """Instantiate a concrete subclass of DebusineNode."""

        class TestNode(DebusineNode):
            @override
            def render_checked(self, _: Context) -> str:
                raise NotImplementedError()

        return TestNode()

    @override_settings(TEST_MODE=False)
    def test_render_error_debug_or_tests(self) -> None:
        node = self.make_node()
        for setting in ("DEBUG", "TEST_MODE"):
            with (
                self.subTest(setting=setting),
                override_settings(**{setting: True}),
                self.assertRaisesRegex(ValueError, r"devel message"),
            ):
                node._render_error(
                    user_message="user message", devel_message="devel message"
                )

    @override_settings(TEST_MODE=False)
    def test_render_error_exception_debug_or_tests(self) -> None:
        node = self.make_node()
        self.enterContext(
            mock.patch.object(
                node,
                "render_checked",
                side_effect=RuntimeError("expected error"),
            )
        )

        for setting in ("DEBUG", "TEST_MODE"):
            with (
                self.subTest(setting=setting),
                override_settings(**{setting: True}),
            ):
                with self.assertRaises(RuntimeError) as exc:
                    node.render(Context())
                self.assertEqual(str(exc.exception), "expected error")
                self.assertRegex(
                    exc.exception.__notes__[0],
                    r"unhandled exception in render_checked",
                )

    @override_settings(TEST_MODE=False)
    def test_render_error_production(self) -> None:
        node = self.make_node()
        with self.assertLogs("debusine.web", level="WARNING") as log:
            rendered = node._render_error(
                user_message="user message", devel_message="devel message"
            )

        tree = self.assertHTMLValid(rendered, check_widget_errors=False)
        span = self.assertHasElement(tree, "body/span")
        self.assertEqual(span.get("data-role"), "debusine-template-error")
        _, actual_message = self.get_node_text_normalized(span).split(": ")
        self.assertEqual(actual_message, "user message")

        self.assertEqual(
            log.output,
            ["WARNING:debusine.web:TestNode rendering error: devel message"],
        )

    def test_resolve_filter(self) -> None:
        parser = Parser("")
        expr = FilterExpression("test", parser)
        node = self.make_node()
        self.assertEqual(
            node.resolve_filter(Context({"test": "test"}), expr, "test"), "test"
        )

    def test_resolve_filter_invalid_filter(self) -> None:
        parser = Parser("")
        expr = FilterExpression("expr", parser)
        node = self.make_node()
        with (
            self.assertRaises(DebusineNodeRenderError) as exc,
            mock.patch(
                "django.template.base.FilterExpression.resolve",
                side_effect=RuntimeError("expected error"),
            ),
        ):
            node.resolve_filter(Context({}), expr, "test")
        self.assertEqual(
            exc.exception.user_message, "template argument malformed"
        )
        self.assertEqual(
            exc.exception.devel_message,
            "Invalid test argument: 'expr'",
        )

    def test_resolve_filter_invalid_lookup(self) -> None:
        parser = Parser("")
        expr = FilterExpression("expr", parser)
        node = self.make_node()
        with self.assertRaises(DebusineNodeRenderError) as exc:
            node.resolve_filter(Context({}), expr, "test")
        self.assertEqual(
            exc.exception.user_message, "template argument lookup failed"
        )
        self.assertEqual(
            exc.exception.devel_message,
            "test argument 'expr' failed to resolve",
        )

    def test_render_checked_debusine_node_render_error(self) -> None:
        node = self.make_node()
        exception = RuntimeError("expected error")
        with (
            mock.patch.object(
                node,
                "render_checked",
                side_effect=DebusineNodeRenderError(
                    user_message="um", devel_message="dm", exception=exception
                ),
            ),
            mock.patch.object(
                node, "_render_error", return_value="result"
            ) as render_error,
        ):
            self.assertEqual(node.render(Context()), "result")

        render_error.assert_called_once_with(
            user_message="um", devel_message="dm", exception=exception
        )

    def test_render_checked_exception(self) -> None:
        node = self.make_node()
        exception = RuntimeError("expected error")
        with (
            mock.patch.object(
                node,
                "render_checked",
                side_effect=exception,
            ),
            mock.patch.object(
                node, "_render_error", return_value="result"
            ) as render_error,
        ):
            self.assertEqual(node.render(Context()), "result")

        render_error.assert_called_once_with(
            user_message="template rendering error",
            devel_message="unhandled exception in render_checked",
            exception=exception,
        )


class WidgetNodeTests(TestCase):
    """Tests for the :py:class:`WidgetNode` class."""

    def test_tag_arg_validation(self) -> None:
        for arg in ("", "foo bar", "foo bar baz"):
            with (
                self.subTest(arg=arg),
                self.assertRaisesRegex(
                    TemplateSyntaxError,
                    re.escape("{% widget %} requires exactly one argument"),
                ),
            ):
                engines["django"].from_string(
                    "{% load debusine %}{% widget " + arg + " %}"
                )

    def test_render(self) -> None:
        request = RequestFactory().get("/")
        template = engines["django"].from_string(
            "{% load debusine %}{% widget 'value' %}"
        )
        self.assertEqual(template.render({}, request=request), "value")

    def test_string_autoescape(self) -> None:
        node = WidgetNode(cast(FilterExpression, None))
        with mock.patch.object(
            node, "resolve_filter", return_value="<resolved>"
        ):
            self.assertEqual(node.render_checked(Context()), "&lt;resolved&gt;")

    def test_string_autoescape_safestring(self) -> None:
        node = WidgetNode(cast(FilterExpression, None))
        with mock.patch.object(
            node, "resolve_filter", return_value=SafeString("<resolved>")
        ):
            self.assertEqual(node.render_checked(Context()), "<resolved>")

    def test_string_noautoescape(self) -> None:
        node = WidgetNode(cast(FilterExpression, None))
        with mock.patch.object(
            node, "resolve_filter", return_value="<resolved>"
        ):
            self.assertEqual(
                node.render_checked(Context(autoescape=False)),
                "<resolved>",
            )

    def test_widget(self) -> None:
        class _Test(Widget):
            @override
            def render(self, _: Context) -> str:
                return "rendered"

        node = WidgetNode(cast(FilterExpression, None))
        with mock.patch.object(node, "resolve_filter", return_value=_Test()):
            self.assertEqual(node.render_checked(Context()), "rendered")

    def test_widget_raises_exception(self) -> None:
        exception = RuntimeError("expected error")

        class _Test(Widget):
            @override
            def render(self, _: Context) -> str:
                raise exception

        node = WidgetNode(cast(FilterExpression, None))
        with (
            mock.patch.object(node, "resolve_filter", return_value=_Test()),
            self.assertRaises(DebusineNodeRenderError) as exc,
        ):
            self.assertEqual(node.render_checked(Context()), "error")
        self.assertEqual(exc.exception.user_message, "_Test failed to render")
        self.assertRegex(
            exc.exception.devel_message,
            r"Widget None \(.+\._Test object at .+\) failed to render",
        )
        self.assertEqual(exc.exception.exception, exception)

    def test_value_errors(self) -> None:
        """Test rendering with problematic values."""
        node = WidgetNode(cast(FilterExpression, None))
        for value, user_message, devel_message in (
            (42, "invalid widget type", "widget None 42 has invalid type"),
            (3.14, "invalid widget type", "widget None 3.14 has invalid type"),
            ({}, "invalid widget type", "widget None {} has invalid type"),
        ):
            with (
                self.subTest(value=value),
                mock.patch.object(node, "resolve_filter", return_value=value),
            ):
                with self.assertRaises(DebusineNodeRenderError) as exc:
                    node.render_checked(Context())
                self.assertEqual(exc.exception.user_message, user_message)
                self.assertEqual(exc.exception.devel_message, devel_message)


class WithscopeTests(TestCase):
    """Tests for withscope tag."""

    scenario = scenarios.DefaultContext()
    scope2: ClassVar[Scope]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up a database layout for views."""
        super().setUpTestData()
        cls.scope2 = cls.playground.get_or_create_scope("scope2")

    def render(self, template_code: str) -> str:
        """Render a template from a string."""
        request = RequestFactory().get("/")
        template = engines["django"].from_string(template_code)
        return template.render({"context": context}, request=request)

    def test_context_accessors_defaults(self) -> None:
        """Test defaults for application context accessors."""
        self.assertEqual(self.render("{{scope.name}}"), "")
        self.assertEqual(self.render("{{workspace.name}}"), "")

    def test_context_accessors_populated(self) -> None:
        """Test application context accessors for populated contexts."""
        self.scenario.set_current()
        self.assertEqual(
            self.render("{{scope.name}}"), self.scenario.scope.name
        )
        self.assertEqual(
            self.render("{{context.user.username}}"),
            self.scenario.user.username,
        )
        self.assertEqual(
            self.render("{{workspace.name}}"), self.scenario.workspace.name
        )

    def test_withscope(self) -> None:
        """Test withscope template tag."""
        self.scenario.set_current()
        name = self.scenario.scope.name
        self.assertEqual(
            self.render(
                "{% load debusine %}{{scope}}"
                "{% withscope 'scope2' %}{{scope}}{% endwithscope %}"
                "{{scope}}"
            ),
            f"{name}scope2{name}",
        )

    def test_withscope_preserve_user(self) -> None:
        self.scenario.set_current()
        self.assertEqual(
            self.render(
                "{% load debusine %}"
                "{% withscope 'scope2' %}"
                "{{scope}}{{context.user}}"
                "{% endwithscope %}"
            ),
            f"scope2{self.scenario.user.username}",
        )

    def test_withscope_misspelled(self) -> None:
        """Test withscope with a misspelled variable."""
        self.scenario.set_current()
        self.assertEqual(
            self.render(
                "{% load debusine %}{{scope}}"
                "{% withscope misspelled_var %}{{scope}}{% endwithscope %}"
                "{{scope}}"
            ),
            self.scenario.scope.name * 3,
        )

    def test_withscope_wrongtype(self) -> None:
        """Test withscope with a scope of an inappropriate type."""
        self.scenario.set_current()
        self.assertEqual(
            self.render(
                "{% load debusine %}{{scope}}"
                "{% withscope 3 %}{{scope}}{% endwithscope %}"
                "{{scope}}"
            ),
            self.scenario.scope.name * 3,
        )

    def test_withscope_wrongscope(self) -> None:
        """Test withscope with a nonexistent scope."""
        self.scenario.set_current()
        self.assertEqual(
            self.render(
                "{% load debusine %}{{scope}}"
                "{% withscope 'wrongscope' %}{{scope}}{% endwithscope %}"
                "{{scope}}"
            ),
            self.scenario.scope.name * 3,
        )

    def test_withscope_noarg(self) -> None:
        """Test withscope without args."""
        with self.assertRaisesRegex(
            TemplateSyntaxError, "withscope requires exactly one argument"
        ):
            self.render(
                "{% load debusine %}{% withscope %}{{scope}}{% endwithscope %}"
            )

    def test_withscope_toomanyargs(self) -> None:
        """Test withscope with too many args."""
        with self.assertRaisesRegex(
            TemplateSyntaxError, "withscope requires exactly one argument"
        ):
            self.render(
                "{% load debusine %}"
                "{% withscope a b %}{{scope}}{% endwithscope %}"
            )


class MockResource:
    """Mock a permission predicate for has_perm tests."""

    user: PermissionUser | None

    def __init__(self, retval: bool) -> None:
        """Store the return value."""
        self.retval = retval

    def predicate(self, pc: PermissionContext) -> bool:
        """Store the argument for checking later."""
        self.user = pc.user
        return self.retval

    def get_roles(self, pc: PermissionContext) -> list[str]:
        """Get the roles of the user."""
        # Mock as roles the user name and an arbitrary name, since get_roles
        # can return multiple values
        return [str(pc.user), "a_test_role"]


class HaspermTests(TestCase):
    """Test the has_perm template filter."""

    scenario = scenarios.DefaultScopeUser()

    def test_context_worker_token(self) -> None:
        """Test using the default user."""
        token = self.playground.create_worker_token()
        context.set_permission_context_scoped(
            self.scenario.scope, PermissionContext.for_worker(token)
        )

        resource = MockResource(True)
        self.assertTrue(
            has_perm(resource, "predicate"),  # type: ignore[arg-type]
        )
        self.assertEqual(resource.user, AnonymousUser())

    def test_context_user_anonymous(self) -> None:
        """Test using the default user."""
        context.set_permission_context_scoped(
            self.scenario.scope, PermissionContext.for_ui_anonymous()
        )
        resource = MockResource(True)
        self.assertTrue(
            has_perm(resource, "predicate"),  # type: ignore[arg-type]
        )
        assert resource.user is not None
        self.assertFalse(resource.user.is_authenticated)

    def test_context_user(self) -> None:
        """Test using the default user."""
        self.scenario.set_current()
        resource = MockResource(False)
        self.assertFalse(
            has_perm(resource, "predicate"),  # type: ignore[arg-type]
        )
        assert resource.user is not None
        self.assertEqual(resource.user, self.scenario.user)


class RolesTests(TestCase):
    """Test the roles filter."""

    scenario = scenarios.DefaultScopeUser()

    def test_current(self) -> None:
        """Test with the default user."""
        self.scenario.set_current()

        resource = cast(Model, MockResource(False))
        self.assertEqual(roles(resource), ["a_test_role", "playground"])

    def test_explicit_anonymous(self) -> None:
        """Test passing an anonymous user explicitly."""
        self.scenario.set_current()
        resource = cast(Model, MockResource(False))
        self.assertEqual(
            roles(resource, AnonymousUser()), ["AnonymousUser", "a_test_role"]
        )

    def test_explicit(self) -> None:
        """Test passing a user explicitly."""
        self.scenario.set_current()
        user = self.playground.create_user("other")

        resource = cast(Model, MockResource(False))
        self.assertEqual(roles(resource, user), ["a_test_role", "other"])

    def test_response_without_role(self) -> None:
        self.scenario.set_current()
        resource = cast(Model, object())
        self.assertEqual(roles(resource), [])


class SortedTests(TestCase):
    """Test the sorted template filter."""

    def test_sorted(self) -> None:
        """Test sorting a sequence."""
        self.assertEqual(sorted_([1, 5, 2, 4, 3]), [1, 2, 3, 4, 5])


class HighlightAsYamlTests(TestCase):
    """Test the highlight_as_yaml filter."""

    def test_highlight_as_yaml(self) -> None:
        tree = self.assertHTMLValid(as_yaml(42))

        div = self.assertHasElement(tree, "body/div")
        self.assertEqual(div.get("class"), "formatted-content")
        formatted = div.xpath("string(pre)")
        self.assertEqual(formatted, "42\n...\n")

    def test_highlight_as_yaml_sorted(self) -> None:
        tree = self.assertHTMLValid(as_yaml({"b": 2, "a": 1}))
        formatted = tree.xpath("string(body/div/pre)")
        self.assertEqual(formatted, "a: 1\nb: 2\n")

    def test_highlight_as_yaml_unsorted(self) -> None:
        tree = self.assertHTMLValid(as_yaml({"b": 2, "a": 1}, "unsorted"))
        formatted = tree.xpath("string(body/div/pre)")
        self.assertEqual(formatted, "b: 2\na: 1\n")


class IconTests(TestCase):
    """Test the icon filter."""

    def test_icon(self) -> None:
        for name, expected in (
            ("USER", Icons.USER),
            ("CaTeGoRy", Icons.CATEGORY),
            ("created_at", Icons.CREATED_AT),
            ("does-not-exist", "square-fill"),
        ):
            with self.subTest(name=name):
                self.assertEqual(icon(name), "bi-" + expected)


class HelpTests(TestCase):
    """Test the "help" template tag."""

    def test_help(self) -> None:
        help_data = HELPS["dependencies"]

        html = _help("dependencies")

        tree = self.assertHTMLValid(html)

        self.assertEqual(tree.body.i.attrib["title"], help_data.title)
        self.assertIn(help_data.link, tree.body.i.attrib["data-bs-content"])
        self.assertIn(help_data.summary, tree.body.i.attrib["data-bs-content"])


class UINodeTests(TestCase):
    """Tests for the :py:class:`UINode` class."""

    scenario = scenarios.DefaultScopeUser()

    syntax_error = re.escape(
        "'ui' statement syntax is {% ui [object] as [variable] %}"
    )

    @override
    def setUp(self) -> None:
        super().setUp()
        factory = RequestFactory()
        self.request = factory.get("/")

    def compile(self, template_code: str) -> DjangoTemplate:
        """Compile a template snippet."""
        template = engines["django"].from_string(
            "{% load debusine %}" + template_code
        )
        assert isinstance(template, DjangoTemplate)
        return template

    def test_ui(self) -> None:
        self.compile("{% ui test as 'var' %}")

    def test_ui_no_args(self) -> None:
        with self.assertRaisesRegex(TemplateSyntaxError, self.syntax_error):
            self.compile("{% ui %}")

    def test_ui_no_as(self) -> None:
        with self.assertRaisesRegex(TemplateSyntaxError, self.syntax_error):
            self.compile("{% ui helper %}")

    def test_ui_no_target(self) -> None:
        with self.assertRaisesRegex(TemplateSyntaxError, self.syntax_error):
            self.compile("{% ui as %}")

    def test_ui_multiple_targets(self) -> None:
        with self.assertRaisesRegex(TemplateSyntaxError, self.syntax_error):
            self.compile("{% ui as a b %}")

    def test_lookup_and_set(self) -> None:
        node = UINode(cast(FilterExpression, None), "var")
        with (
            mock.patch.object(
                node, "resolve_filter", return_value=self.scenario.scope
            ),
            mock.patch(
                "debusine.web.views.ui.base.UI.for_instance",
                return_value="result",
            ) as for_instance,
        ):
            context = Context({"request": self.request})
            self.assertEqual(node.render_checked(context), "")
            for_instance.assert_called_once_with(
                self.request, self.scenario.scope
            )
            self.assertEqual(context["var"], "result")

    def test_value_errors(self) -> None:
        """Test lookup with problematic values."""
        node = UINode(cast(FilterExpression, None), "var")
        for value, devel_message in (
            ("value", "ui model instance None 'value' has invalid type"),
            (42, "ui model instance None 42 has invalid type"),
            (3.14, "ui model instance None 3.14 has invalid type"),
        ):
            with (
                self.subTest(value=value),
                mock.patch.object(node, "resolve_filter", return_value=value),
            ):
                context = Context()
                with self.assertRaises(DebusineNodeRenderError) as exc:
                    node.render_checked(context)
                self.assertIsNone(context["var"])

                self.assertEqual(
                    exc.exception.user_message,
                    "template argument lookup failed",
                )
                self.assertEqual(exc.exception.devel_message, devel_message)

    def test_model_without_helper(self) -> None:
        node = UINode(cast(FilterExpression, None), "var")
        exception = KeyError("not found")
        with (
            mock.patch.object(
                node, "resolve_filter", return_value=self.scenario.scope
            ),
            mock.patch(
                "debusine.web.views.ui.base.UI.for_instance",
                side_effect=exception,
            ) as for_instance,
        ):
            context = Context({"request": self.request})
            with self.assertRaises(DebusineNodeRenderError) as exc:
                node.render_checked(context)
            for_instance.assert_called_once_with(
                self.request, self.scenario.scope
            )
            self.assertIsNone(context["var"])

            self.assertEqual(
                exc.exception.user_message,
                "template argument lookup failed",
            )
            self.assertEqual(
                exc.exception.devel_message,
                "ui helper lookup for <Scope: debusine> failed",
            )
            self.assertEqual(exc.exception.exception, exception)


class PlaceNodeTests(TestCase):
    """Tests for the :py:class:`PlaceNode` class."""

    scenario = scenarios.DefaultScopeUser()

    usage = (
        "Usage: {% place {object} [place_type] as_{widget_type}"
        " [key=value, ...] %}"
    )

    @override
    def setUp(self) -> None:
        super().setUp()
        factory = RequestFactory()
        self.request = factory.get("/")

    def compile(self, template_code: str) -> DjangoTemplate:
        """Compile a template snippet."""
        template = engines["django"].from_string(
            "{% load debusine %}" + template_code
        )
        assert isinstance(template, DjangoTemplate)
        return template

    def get_node(self, template_str: str) -> PlaceNode:
        """Return the first PlaceNode in the compiled template."""
        template = self.compile(template_str)
        nodelist = template.template.nodelist
        for node in nodelist:
            if isinstance(node, PlaceNode):
                return node
        # This is currently never reached, but better have it as a safety net
        # than removing it to get to 100% coverage
        self.fail(
            f"{template}: nodelist does not contain a PlaceNode: {nodelist}"
        )  # pragma: no cover

    def assertCompileFails(self, template: str, error: str) -> None:
        with self.assertRaisesRegex(
            TemplateSyntaxError, re.escape(error + ". " + self.usage)
        ):
            self.compile(template)

    def test_tag_without_place_type(self) -> None:
        node = self.get_node("{% place test as_button %}")
        self.assertIsNone(node.place_type)
        self.assertEqual(node.widget_type, "as_button")

    def test_tag_with_place_type(self) -> None:
        node = self.get_node("{% place test foo as_button %}")
        self.assertEqual(node.place_type, "foo")
        self.assertEqual(node.widget_type, "as_button")

    def test_tag_without_place_type_with_kwargs(self) -> None:
        node = self.get_node("{% place test as_button k=3 %}")
        self.assertIsNone(node.place_type)
        self.assertEqual(node.widget_type, "as_button")
        self.assertEqual(node.kwargs[0][0], "k")

    def test_tag_with_place_type_with_kwargs(self) -> None:
        node = self.get_node("{% place test foo as_button k=3 v=5 %}")
        self.assertEqual(node.place_type, "foo")
        self.assertEqual(node.widget_type, "as_button")
        self.assertEqual(node.kwargs[0][0], "k")
        self.assertEqual(node.kwargs[1][0], "v")

    def test_tag_no_args(self) -> None:
        for template, error in (
            ("{% place %}", "object not provided"),
            ("{% place obj %}", "as_[widget] not provided"),
            ("{% place obj ptype %}", "as_[widget] not provided"),
            (
                "{% place obj ptype foo %}",
                "widget type does not begin with `as_`",
            ),
            ("{% place obj as_foo k %}", "missing = in key=value assignment"),
            ("{% place obj as_foo = %}", "empty key in key=value assignment"),
            (
                "{% place obj as_foo =foo %}",
                "empty key in key=value assignment",
            ),
        ):
            with self.subTest(template=template):
                self.assertCompileFails(template, error)

    def test_resolve_place(self) -> None:
        self.scenario.set_current()
        self.request.user = self.scenario.user
        scope = self.scenario.scope
        ui = scope.ui(self.request)
        for obj, place_type, expected in (
            (scope, None, ui.place),
            (ui, None, ui.place),
            (ui.place, None, ui.place),
            (scope, "group_list", ui.place_group_list),
            (ui, "group_list", ui.place_group_list),
            (ui.place, "group_list", ui.place),
        ):
            with self.subTest(obj=repr(obj), place_type=place_type):
                context = Context({"request": self.request})
                node = PlaceNode(
                    cast(FilterExpression, None), place_type, "as_button", []
                )
                with mock.patch.object(
                    node, "resolve_filter", return_value=obj
                ):
                    self.assertIs(node._resolve_place(context), expected)

    def test_resolve_place_bad_object_type(self) -> None:
        self.scenario.set_current()
        node = PlaceNode(cast(FilterExpression, None), None, "as_button", [])
        with (
            mock.patch.object(node, "resolve_filter", return_value=42),
            self.assertRaises(DebusineNodeRenderError) as exc,
        ):
            node._resolve_place(Context())
        self.assertEqual(exc.exception.user_message, "invalid object type")
        self.assertEqual(
            exc.exception.devel_message,
            "object None resolved to unsupported object type int (42)",
        )

    def test_resolve_place_model_without_helper(self) -> None:
        self.scenario.set_current()
        context = Context({"request": self.request})
        node = PlaceNode(cast(FilterExpression, None), None, "as_button", [])
        exception = KeyError("not found")
        with (
            mock.patch.object(
                node, "resolve_filter", return_value=self.scenario.scope
            ),
            mock.patch(
                "debusine.web.views.ui.base.UI.for_instance",
                side_effect=exception,
            ),
            self.assertRaises(DebusineNodeRenderError) as exc,
        ):
            node._resolve_place(context)
        self.assertEqual(exc.exception.user_message, "ui lookup failed")
        self.assertEqual(
            exc.exception.devel_message, "ui helper lookup for None failed"
        )
        self.assertEqual(exc.exception.exception, exception)

    def test_resolve_place_bad_place_type(self) -> None:
        self.scenario.set_current()
        context = Context({"request": self.request})
        ui = self.scenario.scope.ui(self.request)
        node = PlaceNode(
            cast(FilterExpression, None), "invalid", "as_button", []
        )
        with (
            mock.patch.object(node, "resolve_filter", return_value=ui),
            self.assertRaises(DebusineNodeRenderError) as exc,
        ):
            node._resolve_place(context)
        self.assertEqual(exc.exception.user_message, "place lookup failed")
        self.assertEqual(
            exc.exception.devel_message,
            f"place method 'place_invalid' for {ui!r} failed",
        )
        self.assertIsInstance(exc.exception.exception, AttributeError)

    def test_resolve_kwargs(self) -> None:
        context = Context()
        node = PlaceNode(
            cast(FilterExpression, None),
            None,
            "as_button",
            [
                ("key1", cast(FilterExpression, 1)),
                ("key2", cast(FilterExpression, 2)),
            ],
        )
        with mock.patch.object(
            node, "resolve_filter", return_value="value"
        ) as resolve_filter:
            self.assertEqual(
                node._resolve_kwargs(context),
                {"key1": "value", "key2": "value"},
            )
        resolve_filter.assert_has_calls(
            [
                mock.call(context, 1, "kwarg value"),
                mock.call(context, 2, "kwarg value"),
            ]
        )

    def test_render_no_kwargs(self) -> None:
        ui = self.scenario.scope.ui(self.request)
        node = PlaceNode(cast(FilterExpression, None), None, "as_button", [])
        with (
            mock.patch.object(node, "_resolve_place", return_value=ui.place),
            mock.patch.object(
                ui.place, "as_button", return_value="rendered"
            ) as renderer,
        ):
            self.assertEqual(node.render_checked(Context()), "rendered")

        renderer.assert_called_once_with()

    def test_render_with_kwargs(self) -> None:
        ui = self.scenario.scope.ui(self.request)
        node = PlaceNode(cast(FilterExpression, None), None, "as_button", [])
        with (
            mock.patch.object(node, "_resolve_place", return_value=ui.place),
            mock.patch.object(
                node, "_resolve_kwargs", return_value={"key": "value"}
            ),
            mock.patch.object(
                ui.place, "as_button", return_value="rendered"
            ) as renderer,
        ):
            self.assertEqual(node.render_checked(Context()), "rendered")

        renderer.assert_called_once_with(key="value")

    def test_render_invalid_widget_type(self) -> None:
        ui = self.scenario.scope.ui(self.request)
        node = PlaceNode(cast(FilterExpression, None), None, "as_invalid", [])
        with (
            mock.patch.object(node, "_resolve_place", return_value=ui.place),
            self.assertRaises(DebusineNodeRenderError) as exc,
        ):
            node.render_checked(Context())

        self.assertEqual(
            exc.exception.user_message, "invalid place render type"
        )
        self.assertEqual(
            exc.exception.devel_message,
            f"place {ui.place!r} does not have an as_invalid method",
        )
        self.assertIsNone(exc.exception.exception)

    def test_render_fails(self) -> None:
        exception = RuntimeError("expected error")
        ui = self.scenario.scope.ui(self.request)
        node = PlaceNode(cast(FilterExpression, None), None, "as_button", [])
        with (
            mock.patch.object(node, "_resolve_place", return_value=ui.place),
            mock.patch.object(ui.place, "as_button", side_effect=exception),
            self.assertRaises(DebusineNodeRenderError) as exc,
        ):
            node.render_checked(Context())

        self.assertEqual(exc.exception.user_message, "place rendering failed")
        self.assertEqual(
            exc.exception.devel_message,
            f"place {ui.place!r} as_button method failed",
        )
        self.assertIs(exc.exception.exception, exception)


class MessageToastTests(TestCase):
    """Test the message_toast* helper filters."""

    def test_color_class(self) -> None:
        for level, expected_class in (
            (messages.DEBUG, "light"),
            (messages.INFO, "light"),
            (messages.SUCCESS, "success"),
            (messages.WARNING, "warning"),
            (messages.ERROR, "danger"),
            (12345, "light"),
        ):
            with self.subTest(level=level):
                msg = Message(level, "test")
                self.assertEqual(message_toast_color_class(msg), expected_class)

    def test_header_icon(self) -> None:
        for level, expected_icon in (
            (messages.DEBUG, "bi bi-info-circle"),
            (messages.INFO, "bi bi-info-circle"),
            (messages.SUCCESS, "bi bi-info-circle"),
            (messages.WARNING, "bi bi-exclamation-triangle"),
            (messages.ERROR, "bi bi-exclamation-triangle"),
            (12345, "bi bi-info-circle"),
        ):
            with self.subTest(level=level):
                msg = Message(level, "test")
                self.assertEqual(message_toast_header_icon(msg), expected_icon)

    def test_worker_last_seen_never(self) -> None:
        self.assertEqual(worker_last_seen(None), "Never")

    def test_worker_last_seen_less_than_minutes_ago(self) -> None:
        actual = worker_last_seen(dt.datetime.now(dt.UTC))
        minutes = int(TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL.total_seconds() / 60)
        self.assertEqual(actual, f"Less than {minutes} minutes ago")

    def test_worker_last_seen_ago(self) -> None:
        actual = worker_last_seen(
            dt.datetime.now(dt.UTC)
            - TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL
            - dt.timedelta(seconds=1)
        )

        self.assertRegex(actual, " ago$")
        self.assertNotIn("Less than", actual)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for queue and worker pool summaries."""

from typing import ClassVar, override

from django.test import override_settings

from debusine.artifacts.models import TaskTypes
from debusine.db.models import Artifact, WorkRequest, Worker
from debusine.server.status import (
    QueueStatus,
    TaskQueueSummary,
    WorkerStatus,
    WorkerSummary,
)
from debusine.server.tasks.wait.models import ExternalDebsignData
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.tasks.models import (
    MmDebstrapBootstrapOptions,
    MmDebstrapData,
    SystemBootstrapRepository,
    WorkerType,
)
from debusine.test.django import TestCase


# We want to be in control of the scheduling status of tasks, for these tests
@override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
class WorkerStatusTests(TestCase):
    """Tests for WorkerStatus."""

    source: ClassVar[Artifact]
    environment: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source = cls.playground.create_minimal_source_package_artifact()
        environment = cls.playground.create_debian_environment()
        assert environment.artifact is not None
        cls.environment = environment.artifact

    def create_worker_task(
        self,
        architecture: str | None = None,
        *,
        mark_pending: bool = False,
        mark_running: bool = False,
        worker: Worker | None = None,
    ) -> WorkRequest:
        """Create a worker task for the given architecture."""
        if architecture is None:
            return self.playground.create_worker_task(
                task_name="noop",
                mark_pending=mark_pending,
                mark_running=mark_running,
                worker=worker,
            )
        else:
            work_request = self.playground.create_sbuild_work_request(
                source=self.source,
                environment=self.environment,
                architecture=architecture,
            )
            self.playground.advance_work_request(
                work_request,
                mark_pending=mark_pending,
                mark_running=mark_running,
                worker=worker,
            )
            return work_request

    def assertTasks(
        self, status: QueueStatus, task_type: TaskTypes, state: str, value: int
    ) -> None:
        """Assert tasks in status for task_type have state == value."""
        type_tasks = status.tasks.get(task_type, TaskQueueSummary())
        self.assertEqual(getattr(type_tasks, state), value)

    def assertWorkers(
        self,
        status: QueueStatus,
        worker_type: WorkerType,
        state: str,
        value: int,
    ) -> None:
        """Assert workers in status for worker_type have state == value."""
        type_status = status.workers.get(worker_type, WorkerSummary())
        self.assertEqual(getattr(type_status, state), value)

    def assertWorkerTasks(
        self,
        status: QueueStatus,
        worker_type: WorkerType,
        state: str,
        value: int,
    ) -> None:
        """Assert worker_tasks in status for worker_type have state == value."""
        type_tasks = status.worker_tasks.get(worker_type, TaskQueueSummary())
        self.assertEqual(getattr(type_tasks, state), value)

    def assertExternalWorkersArch(
        self, status: QueueStatus, arch: str, state: str, value: int
    ) -> None:
        """Assert external_workers_arch for arch have state == value."""
        arch_status = status.external_workers_arch.get(arch, WorkerSummary())
        self.assertEqual(getattr(arch_status, state), value)

    def assertExternalWorkersNativeArch(
        self, status: QueueStatus, arch: str, state: str, value: int
    ) -> None:
        """Assert external_workers_native_arch for arch have state == value."""
        arch_status = status.external_workers_native_arch.get(
            arch, WorkerSummary()
        )
        self.assertEqual(getattr(arch_status, state), value)

    def assertWorkerTasksArch(
        self, status: QueueStatus, arch: str, state: str, value: int
    ) -> None:
        """Assert worker_tasks_arch in status for arch have state == value."""
        arch_tasks = status.worker_tasks_arch.get(arch, TaskQueueSummary())
        self.assertEqual(getattr(arch_tasks, state), value)

    def test_counts_tasks_by_type(self) -> None:
        """Test that a tasks counts tasks by type."""
        self.create_worker_task(mark_pending=True)
        self.playground.create_signing_task(task_name="noop", mark_running=True)

        status = WorkerStatus.get_status()
        self.assertTasks(status, TaskTypes.WORKER, "pending", 1)
        self.assertTasks(status, TaskTypes.WORKER, "running", 0)
        self.assertTasks(status, TaskTypes.SIGNING, "pending", 0)
        self.assertTasks(status, TaskTypes.SIGNING, "running", 1)
        self.assertTasks(status, TaskTypes.INTERNAL, "pending", 0)
        self.assertTasks(status, TaskTypes.INTERNAL, "running", 0)
        self.assertWorkerTasks(status, WorkerType.EXTERNAL, "pending", 1)
        self.assertWorkerTasks(status, WorkerType.EXTERNAL, "running", 0)
        self.assertWorkerTasks(status, WorkerType.SIGNING, "pending", 0)
        self.assertWorkerTasks(status, WorkerType.SIGNING, "running", 1)
        self.assertWorkerTasks(status, WorkerType.CELERY, "pending", 0)
        self.assertWorkerTasks(status, WorkerType.CELERY, "running", 0)

    def test_worker_counts_against_all_archs(self) -> None:
        """Test that a worker counts against all architectures it supports."""
        worker = self.playground.create_worker(
            worker_type=WorkerType.EXTERNAL,
            extra_dynamic_metadata={
                "system:native_architecture": "amd64",
                "system:architectures": ["amd64", "i386"],
            },
        )
        worker.mark_connected()

        status = WorkerStatus.get_status()
        self.assertExternalWorkersArch(status, "amd64", "connected", 1)
        self.assertExternalWorkersArch(status, "i386", "connected", 1)
        # and by native_arch:
        self.assertExternalWorkersNativeArch(status, "amd64", "connected", 1)
        self.assertExternalWorkersNativeArch(status, "i386", "connected", 0)
        # and by type:
        self.assertWorkers(status, WorkerType.EXTERNAL, "connected", 1)

    def test_work_request_filters_by_task_type(self) -> None:
        """Test that worker_tasks_arch only includes WORKER work requests."""

    def test_work_request_counts_build_architecture_method(self) -> None:
        """
        Test that a work request is counted against its build_architecture.

        For a WorkRequest with build_architecture not specified in task_data.
        Instead the Tasks's .build_architecture() method is called to determine
        the architecture.
        """
        task_data = MmDebstrapData(
            bootstrap_options=MmDebstrapBootstrapOptions(architecture="amd64"),
            bootstrap_repositories=[
                SystemBootstrapRepository(
                    mirror="https://deb.debian.org/debian", suite="stable"
                )
            ],
        )
        expected_architecture = task_data.bootstrap_options.architecture

        self.playground.create_worker_task(
            task_name="mmdebstrap", task_data=task_data, mark_pending=True
        )
        self.playground.create_worker_task(
            task_name="mmdebstrap", task_data=task_data, mark_running=True
        )

        status = WorkerStatus.get_status()
        self.assertWorkerTasksArch(status, expected_architecture, "pending", 1)
        self.assertWorkerTasksArch(status, expected_architecture, "running", 1)
        self.assertTasks(status, TaskTypes.WORKER, "pending", 1)
        self.assertTasks(status, TaskTypes.WORKER, "running", 1)
        self.assertWorkerTasks(status, WorkerType.EXTERNAL, "pending", 1)
        self.assertWorkerTasks(status, WorkerType.EXTERNAL, "running", 1)

    def test_work_request_non_architecture_constrained(self) -> None:
        """
        Test that a work request without an architecture counts as arch:None.

        Work request without "build_architecture" in task_data and
        build_architecture() returns None.
        """
        self.create_worker_task(mark_pending=True)
        self.create_worker_task(mark_running=True)

        status = WorkerStatus.get_status()

        actual = status.worker_tasks_arch
        expected = {None: TaskQueueSummary(pending=1, running=1)}
        self.assertEqual(actual, expected)

        # Counts by type too:
        self.assertTasks(status, TaskTypes.WORKER, "pending", 1)
        self.assertTasks(status, TaskTypes.WORKER, "running", 1)
        self.assertWorkerTasks(status, WorkerType.EXTERNAL, "pending", 1)
        self.assertWorkerTasks(status, WorkerType.EXTERNAL, "running", 1)

    def test_get_status_no_pending_work_requests(self) -> None:
        """
        get_status: no pending work requests.

        Empty dictionary because no pending work requests and no workers.
        """
        status = WorkerStatus.get_status()
        self.assertEqual(status.external_workers_arch, {})
        self.assertEqual(status.worker_tasks_arch, {})
        for task_type in TaskTypes:
            self.assertEqual(status.tasks[task_type], TaskQueueSummary())
        for worker_type in WorkerType:
            if worker_type == WorkerType.SCHEDULER:
                self.assertNotIn(worker_type, status.workers)
                self.assertNotIn(worker_type, status.worker_tasks)
            else:
                self.assertEqual(status.workers[worker_type], WorkerSummary())
                self.assertEqual(
                    status.worker_tasks[worker_type], TaskQueueSummary()
                )

    def test_disabled_workers_ignored(self) -> None:
        """Workers are not counted if they are disabled."""
        self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arch1",
                "system:architectures": ["arch1", "arch2"],
            }
        )

        worker = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arch3",
                "system:architectures": ["arch3"],
            }
        )
        assert worker.token is not None
        worker.token.disable()

        status = WorkerStatus.get_status()
        self.assertNotIn("arch3", status.external_workers_arch)
        self.assertNotIn("arch3", status.external_workers_native_arch)
        self.assertWorkers(status, WorkerType.EXTERNAL, "registered", 1)

    def test_count_registered_workers(self) -> None:
        """Registered and connected workers are counted."""
        worker = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arch1",
                "system:architectures": ["arch1", "arch2"],
            }
        )
        worker.mark_connected()

        worker = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arch3",
                "system:architectures": ["arch3"],
            }
        )

        status = WorkerStatus.get_status()
        self.assertExternalWorkersArch(status, "arch1", "registered", 1)
        self.assertExternalWorkersArch(status, "arch2", "registered", 1)
        self.assertExternalWorkersArch(status, "arch3", "registered", 1)
        self.assertExternalWorkersArch(status, "arch1", "connected", 1)
        self.assertExternalWorkersArch(status, "arch2", "connected", 1)
        self.assertExternalWorkersArch(status, "arch3", "connected", 0)
        # and by native_arch:
        self.assertExternalWorkersNativeArch(status, "arch1", "registered", 1)
        self.assertExternalWorkersNativeArch(status, "arch2", "registered", 0)
        self.assertExternalWorkersNativeArch(status, "arch3", "registered", 1)
        self.assertExternalWorkersNativeArch(status, "arch1", "connected", 1)
        self.assertExternalWorkersNativeArch(status, "arch2", "connected", 0)
        self.assertExternalWorkersNativeArch(status, "arch3", "connected", 0)
        # and by type:
        self.assertWorkers(status, WorkerType.EXTERNAL, "registered", 2)
        self.assertWorkers(status, WorkerType.EXTERNAL, "connected", 1)

    def test_count_registered_workers_by_type(self) -> None:
        """Registered workers are counted by type."""
        self.playground.create_worker(
            worker_type=WorkerType.EXTERNAL
        ).mark_connected()
        self.playground.create_worker(worker_type=WorkerType.EXTERNAL)
        self.playground.create_worker(
            worker_type=WorkerType.SIGNING
        ).mark_connected()
        self.playground.create_worker(
            worker_type=WorkerType.CELERY
        ).mark_connected()

        status = WorkerStatus.get_status()
        self.assertWorkers(status, WorkerType.EXTERNAL, "registered", 2)
        self.assertWorkers(status, WorkerType.SIGNING, "registered", 1)
        self.assertWorkers(status, WorkerType.CELERY, "registered", 1)
        self.assertWorkers(status, WorkerType.EXTERNAL, "connected", 1)
        self.assertWorkers(status, WorkerType.SIGNING, "connected", 1)
        self.assertWorkers(status, WorkerType.CELERY, "connected", 1)

    def test_count_busy_workers_running(self) -> None:
        """Assigned RUNNING tasks count as a busy worker."""
        worker = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arch1",
                "system:architectures": ["arch1", "arch2"],
            }
        )
        worker.mark_connected()

        self.create_worker_task("arch1", worker=worker, mark_running=True)

        status = WorkerStatus.get_status()
        self.assertExternalWorkersArch(status, "arch1", "busy", 1)
        # the arch2 slot is also busy:
        self.assertExternalWorkersArch(status, "arch2", "busy", 1)
        # native_arch is busy:
        self.assertExternalWorkersNativeArch(status, "arch1", "busy", 1)
        # and by type:
        self.assertWorkers(status, WorkerType.EXTERNAL, "busy", 1)

    def test_count_busy_workers_pending(self) -> None:
        """Assigned PENDING tasks count as a busy worker."""
        worker = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arch1",
                "system:architectures": ["arch1", "arch2"],
            }
        )
        worker.mark_connected()

        self.create_worker_task("arch1", mark_pending=True, worker=worker)

        status = WorkerStatus.get_status()
        self.assertExternalWorkersArch(status, "arch1", "busy", 1)
        # the arch2 slot is also busy:
        self.assertExternalWorkersArch(status, "arch2", "busy", 1)
        # build_arch is busy:
        self.assertExternalWorkersNativeArch(status, "arch1", "busy", 1)
        # and by type:
        self.assertWorkers(status, WorkerType.EXTERNAL, "busy", 1)

    def test_count_idle_workers(self) -> None:
        """Connected workers without assigned tasks are idle."""
        worker = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arch1",
                "system:architectures": ["arch1"],
            }
        )
        worker.mark_connected()
        worker = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arch2",
                "system:architectures": ["arch2"],
            }
        )
        status = WorkerStatus.get_status()
        self.assertExternalWorkersArch(status, "arch1", "idle", 1)
        self.assertExternalWorkersArch(status, "arch2", "registered", 1)
        self.assertExternalWorkersArch(status, "arch2", "idle", 0)
        # build_arch should be identical:
        self.assertExternalWorkersNativeArch(status, "arch1", "idle", 1)
        self.assertExternalWorkersNativeArch(status, "arch2", "registered", 1)
        self.assertExternalWorkersNativeArch(status, "arch2", "idle", 0)
        # and by type:
        self.assertWorkers(status, WorkerType.EXTERNAL, "registered", 2)
        self.assertWorkers(status, WorkerType.EXTERNAL, "idle", 1)

    def test_external_workers_arch_populated_for_all_task_archs(self) -> None:
        """We populate external_workers_arch for all queued architectures."""
        self.create_worker_task("arch1", mark_pending=True)
        status = WorkerStatus.get_status()
        self.assertIn("arch1", status.external_workers_arch)

    def test_worker_tasks_populated_for_all_worker_archs(self) -> None:
        """We populate worker_tasks_arch for all known worker architectures."""
        self.playground.create_worker(
            extra_dynamic_metadata={"system:architectures": ["arch1"]}
        )
        status = WorkerStatus.get_status()
        self.assertIn("arch1", status.worker_tasks_arch)

    def test_missing_system_architecture(self) -> None:
        """Workers without system:architecture log a warning."""
        self.playground.create_worker().mark_connected()
        with self.assertLogs("debusine.server.status") as cm:
            WorkerStatus.get_status()
        self.assertEqual(
            cm.output,
            [
                "WARNING:debusine.server.status:Worker computer-lan missing "
                "system:native_architecture"
            ],
        )

    def test_full_example(self) -> None:
        """Test the full data output of a non-trivial example state."""
        # Create three work requests of type sbuild
        self.create_worker_task("amd64", mark_pending=True)
        self.create_worker_task("amd64", mark_pending=True)
        self.create_worker_task("arm64", mark_pending=True)

        # Create three connected workers
        worker_amd64_1 = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "amd64",
                "system:architectures": ["amd64", "i386"],
            }
        )
        worker_amd64_1.mark_connected()

        worker_arm64_1 = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arm64",
                "system:architectures": ["arm64"],
            }
        )
        worker_arm64_1.mark_connected()

        worker_arm64_2 = self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arm64",
                "system:architectures": ["arm64"],
            }
        )
        worker_arm64_2.mark_connected()

        # Create non-connected worker (will not appear as available
        # worker in the count for arm64)
        self.playground.create_worker(
            extra_dynamic_metadata={
                "system:native_architecture": "arm64",
                "system:architectures": ["arm64"],
            }
        )
        # And a celery worker
        self.playground.create_worker(
            worker_type=WorkerType.CELERY,
        ).mark_connected()
        # And a scheduler worker, which should not appear in the status
        self.playground.create_worker(
            worker_type=WorkerType.SCHEDULER
        ).mark_connected()

        # Create a running task
        self.create_worker_task(
            "arm64", mark_running=True, worker=worker_arm64_1
        )

        # Create a running wait task
        unsigned = self.playground.create_upload_artifacts().upload
        self.playground.create_wait_task(
            task_name="externaldebsign",
            task_data=ExternalDebsignData(unsigned=unsigned.id),
            workflow_data=WorkRequestWorkflowData(needs_input=True),
            mark_running=True,
        )

        actual = WorkerStatus.get_status()

        expected = QueueStatus(
            tasks={
                TaskTypes.WORKER: TaskQueueSummary(pending=3, running=1),
                TaskTypes.SERVER: TaskQueueSummary(),
                TaskTypes.INTERNAL: TaskQueueSummary(),
                TaskTypes.WORKFLOW: TaskQueueSummary(),
                TaskTypes.SIGNING: TaskQueueSummary(),
                TaskTypes.WAIT: TaskQueueSummary(running=1),
            },
            workers={
                WorkerType.CELERY: WorkerSummary(
                    registered=1, connected=1, idle=1
                ),
                WorkerType.EXTERNAL: WorkerSummary(
                    registered=4, connected=3, idle=2, busy=1
                ),
                WorkerType.SIGNING: WorkerSummary(),
            },
            worker_tasks={
                WorkerType.CELERY: TaskQueueSummary(),
                WorkerType.EXTERNAL: TaskQueueSummary(pending=3, running=1),
                WorkerType.SIGNING: TaskQueueSummary(),
            },
            external_workers_arch={
                "amd64": WorkerSummary(
                    registered=1, connected=1, idle=1, busy=0
                ),
                "arm64": WorkerSummary(
                    registered=3, connected=2, idle=1, busy=1
                ),
                "i386": WorkerSummary(
                    registered=1, connected=1, idle=1, busy=0
                ),
            },
            external_workers_native_arch={
                "amd64": WorkerSummary(
                    registered=1, connected=1, idle=1, busy=0
                ),
                "arm64": WorkerSummary(
                    registered=3, connected=2, idle=1, busy=1
                ),
            },
            worker_tasks_arch={
                "amd64": TaskQueueSummary(pending=2, running=0),
                "arm64": TaskQueueSummary(pending=1, running=1),
                "i386": TaskQueueSummary(pending=0, running=0),
            },
        )

        self.assertEqual(actual, expected)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the test playground functions."""

from typing import ClassVar, override

import debusine.db.playground
from debusine.db.context import context
from debusine.db.models import Artifact, FileInArtifact, FileStore, Workspace
from debusine.db.playground import scenarios
from debusine.test.django import (
    PlaygroundTestCase,
    TestCase,
    TransactionTestCase,
)
from debusine.test.playground import Playground


class IntScenario(scenarios.Scenario):
    """Scenario used to test modification isolation between test cases."""

    value: int

    @override
    def build(self, playground: debusine.db.playground.Playground) -> None:
        super().build(playground)
        self.value = 0


class ScenarioTestCase(PlaygroundTestCase):
    """Common scenario-related tests."""

    scenario = scenarios.DefaultContext()
    int_scenario = IntScenario()

    def test_scenario_instantiated(self) -> None:
        """Test that the scenario has been instantiated."""
        self.assertIsNot(self.scenario, self.__class__.scenario)
        self.assertEqual(
            getattr(self, "playground").scenarios["scenario"], self.scenario
        )
        self.assertIsNotNone(self.scenario.scope)
        self.assertIsNone(context.scope)

    def test_scenario_local_changes1(self) -> None:
        """Test local changes isolation between methods."""
        self.assertEqual(self.int_scenario.value, 0)
        self.int_scenario.value = 1

    def test_scenario_local_changes2(self) -> None:
        """Test local changes isolation between methods."""
        self.assertEqual(self.int_scenario.value, 0)
        self.int_scenario.value = 1


class MemoryPlaygroundTestCaseTests(ScenarioTestCase, TestCase):
    """Test instantiation of playground in normal test case."""

    def test_instantiated(self) -> None:
        """Playground is instantiated."""
        self.assertIsInstance(self.playground, Playground)
        self.assertEqual(
            self.playground.file_store.backend, FileStore.BackendChoices.MEMORY
        )


class LocalPlaygroundTestCaseTests(ScenarioTestCase, TestCase):
    """Test instantiation of playground in normal test case."""

    playground_memory_file_store = False
    scenario = scenarios.DefaultContext()

    def test_instantiated(self) -> None:
        """Playground is instantiated."""
        self.assertIsInstance(self.playground, Playground)
        self.assertEqual(
            self.playground.file_store.backend, FileStore.BackendChoices.LOCAL
        )


class NoPlaygroundTestCaseTests(TestCase):
    """Test instantiation of playground in normal test case."""

    playground_needed = False
    scenario = scenarios.DefaultContext()

    def test_instantiated(self) -> None:
        """Playground is instantiated."""
        self.assertFalse(hasattr(self, "playground"))

    def test_scenario_not_instantiated(self) -> None:
        """Test that the scenario has not been instantiated."""
        self.assertFalse(hasattr(self.scenario, "scope"))
        self.assertIsNone(context.scope)


class MemoryPlaygroundTransactionTestCaseTests(
    ScenarioTestCase, TransactionTestCase
):
    """Test instantiation of playground in normal test case."""

    scenario = scenarios.DefaultContext()

    def test_instantiated(self) -> None:
        """Playground is instantiated."""
        self.assertIsInstance(self.playground, Playground)
        self.assertEqual(
            self.playground.file_store.backend, FileStore.BackendChoices.MEMORY
        )


class LocalPlaygroundTransactionTestCaseTests(
    ScenarioTestCase, TransactionTestCase
):
    """Test instantiation of playground in normal test case."""

    playground_memory_file_store = False
    scenario = scenarios.DefaultContext()

    def test_instantiated(self) -> None:
        """Playground is instantiated."""
        self.assertIsInstance(self.playground, Playground)
        self.assertEqual(
            self.playground.file_store.backend, FileStore.BackendChoices.LOCAL
        )


class NoPlaygroundTransactionTestCaseTests(TransactionTestCase):
    """Test instantiation of playground in normal test case."""

    playground_needed = False
    scenario = scenarios.DefaultContext()

    def test_instantiated(self) -> None:
        """Playground is instantiated."""
        self.assertFalse(hasattr(self, "playground"))

    def test_scenario_not_instantiated(self) -> None:
        """Test that the scenario has not been instantiated."""
        self.assertFalse(hasattr(self.scenario, "scope"))
        self.assertIsNone(context.scope)


class _InheritedScenarioTestBase(TestCase):
    scenario1 = scenarios.DefaultContext()


class InheritedScenarioTests(_InheritedScenarioTestBase):
    """Test inheritance of scenario definitions."""

    scenario2 = scenarios.DefaultContext()

    def test_list_playground_scenarios(self) -> None:
        """Test list_playground_scenarios method."""
        self.assertEqual(
            list(_InheritedScenarioTestBase.list_playground_scenarios()),
            [("scenario1", self.__class__.scenario1)],
        )
        self.assertEqual(
            list(self.list_playground_scenarios()),
            [
                ("scenario1", self.__class__.scenario1),
                ("scenario2", self.__class__.scenario2),
            ],
        )

    def test_scenario_instantiated(self) -> None:
        """Test that the scenario has been instantiated."""
        self.assertIsNotNone(self.scenario1.scope)
        self.assertIsNotNone(self.scenario2.scope)
        self.assertIsNone(context.scope)


class SetCurrentScenarioTestCaseTests(TestCase):
    """Test scenario annotations."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_scenario_instantiated(self) -> None:
        """Test that the scenario has been instantiated."""
        self.assertEqual(context.scope, self.scenario.scope)


class SetCurrentScenarioTransactionTestCaseTests(TransactionTestCase):
    """Test scenario annotations."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_scenario_instantiated(self) -> None:
        """Test that the scenario has been instantiated."""
        self.assertEqual(context.scope, self.scenario.scope)


class PlaygroundTransactionalTestMixin(TestCase):
    """Test restoring state across test methods."""

    # This is a common set of tests for both the in-memory store case and the
    # local file store case. This class is deleted at the end of the module to
    # avoid it being run as unittests

    # See https://docs.python.org/3/library/unittest.html#organizing-test-code
    # "The order in which the various tests will be run is determined by
    # sorting the test method names with respect to the built-in ordering for
    # strings."

    workspace: ClassVar[Workspace]
    art1: ClassVar[Artifact]
    art2: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Create a playground to setup test data."""
        super().setUpTestData()
        cls.workspace = cls.playground.get_default_workspace()
        cls.art1, _ = cls.playground.create_artifact(
            {"test.deb": b"testdeb"}, create_files=True
        )
        cls.art2, _ = cls.playground.create_artifact(
            {"test.dsc": b"testdsc", "test.tar.xz": b"testtarxz"},
            create_files=True,
        )

    def assertTestFixtureUnchanged(self) -> None:
        """Check that the text fixture is as originally set up."""
        self.assertEqual(self.playground.default_username, "playground")
        files1 = list(self.art1.files.order_by("id"))
        files2 = list(self.art2.files.order_by("id"))
        self.assertEqual(len(files1), 1)
        self.assertEqual(len(files2), 2)
        self.assertEqual(Artifact.objects.count(), 2)
        backend = self.workspace.scope.download_file_backend(files1[0])
        with backend.get_stream(files1[0]) as fd:
            self.assertEqual(fd.read(), b"testdeb")
        with backend.get_stream(files2[0]) as fd:
            self.assertEqual(fd.read(), b"testdsc")
        with backend.get_stream(files2[1]) as fd:
            self.assertEqual(fd.read(), b"testtarxz")

    def test_alter_file_storage_1(self) -> None:
        """First part of testing transactional file storage behaviour."""
        self.assertTestFixtureUnchanged()
        self.playground.default_username = "changed_username"
        self.assertEqual(self.playground.default_username, "changed_username")

        # Add a new artifact
        self.playground.create_artifact(
            paths={"testfile": b"testdata"}, create_files=True
        )

        # Add a file to an existing artifact
        fileobj = self.playground.create_file_in_backend(
            contents=b"newfilecontents"
        )
        FileInArtifact.objects.create(
            artifact=self.art1, path="newfile", file=fileobj, complete=True
        )

        # Remove a file
        backend = self.playground.get_default_file_store().get_backend_object()
        backend.remove_file(self.art1.files.earliest("id"))

    def test_alter_file_storage_2(self) -> None:
        """Second part of testing transactional file storage behaviour."""
        self.assertTestFixtureUnchanged()


class PlaygroundTransactionalTestMemoryStore(
    PlaygroundTransactionalTestMixin,
):
    """Test transactional behaviour with in-memory file store."""


class PlaygroundTransactionalTestFileStore(PlaygroundTransactionalTestMixin):
    """Test transactional behaviour with on-disk file store."""

    playground_memory_file_store = False


# Avoid running tests from the common base
del ScenarioTestCase
del PlaygroundTransactionalTestMixin



#  Copyright © The Debusine Developers
#  See the AUTHORS file at the top-level directory of this distribution
#
#  This file is part of Debusine. It is subject to the license terms
#  in the LICENSE file found in the top-level directory of this
#  distribution. No part of Debusine, including this file, may be copied,
#  modified, propagated, or distributed except according to the terms
#  contained in the LICENSE file.

"""Package with test-helper code and utilities."""

from debusine.test.base import TestCase

__all__ = [
    "TestCase",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the worker models."""

import datetime as dt
from typing import ClassVar, override
from unittest import mock

from django.db import IntegrityError
from django.test import RequestFactory, override_settings
from django.utils import timezone

from debusine.db.models import Token, WorkRequest, Worker, WorkerPool
from debusine.db.models.workers import WorkerManager
from debusine.tasks.models import WorkerType
from debusine.test.django import TestCase


class WorkerManagerTests(TestCase):
    """Tests for the WorkerManager."""

    def test_connected(self) -> None:
        """WorkerManager.connected() return the connected Workers."""
        workers = {
            name: Worker.objects.create_with_fqdn(
                name,
                token=self.playground.create_worker_token(set_worker=False),
                worker_type=worker_type,
            )
            for name, worker_type in (
                ("connected-celery", WorkerType.CELERY),
                ("connected-no-dynamic-metadata", WorkerType.EXTERNAL),
                ("connected-old-dynamic-metadata", WorkerType.EXTERNAL),
                ("connected-recent-dynamic-metadata", WorkerType.EXTERNAL),
                ("not-connected", WorkerType.EXTERNAL),
                ("not-connected-celery", WorkerType.CELERY),
            )
        }

        workers["connected-celery"].mark_connected()
        workers["connected-no-dynamic-metadata"].mark_connected()
        workers["connected-old-dynamic-metadata"].mark_connected()
        workers["connected-old-dynamic-metadata"].set_dynamic_metadata({})
        workers[
            "connected-old-dynamic-metadata"
        ].dynamic_metadata_updated_at = timezone.now() - dt.timedelta(
            hours=1, minutes=6
        )
        workers["connected-old-dynamic-metadata"].save()
        workers["connected-recent-dynamic-metadata"].mark_connected()
        workers["connected-recent-dynamic-metadata"].set_dynamic_metadata({})

        self.assertQuerySetEqual(
            Worker.objects.connected(),
            [
                workers["connected-celery"],
                workers["connected-recent-dynamic-metadata"],
            ],
            ordered=False,
        )

    def test_active(self) -> None:
        """Test active() filter."""
        worker_pool = self.playground.create_worker_pool()
        workers: dict[str, Worker] = {}

        workers["static_disconnected"] = self.playground.create_worker(
            fqdn="disconnected.static.lan"
        )

        workers["static_connected_idle"] = self.playground.create_worker(
            fqdn="idle.connected.static.lan"
        )

        workers["dynamic_terminated"] = self.playground.create_worker(
            fqdn="disconnected.dynamic.lan", worker_pool=worker_pool
        )
        workers["dynamic_terminated"].instance_created_at = None
        workers["dynamic_terminated"].save()

        workers["dynamic_disconnected"] = self.playground.create_worker(
            fqdn="disconnected.dynamic.lan", worker_pool=worker_pool
        )

        workers["dynamic_connected_idle"] = self.playground.create_worker(
            fqdn="idle.connected.dynamic.lan", worker_pool=worker_pool
        )
        workers["dynamic_connected_idle"].connected_at = timezone.now()
        workers["dynamic_connected_idle"].save()

        workers["dynamic_connected_busy"] = self.playground.create_worker(
            fqdn="busy.connected.dynamic.lan", worker_pool=worker_pool
        )
        workers["dynamic_connected_busy"].connected_at = timezone.now()
        workers["dynamic_connected_busy"].save()
        wr = self.playground.create_worker_task(mark_pending=True)
        wr.assign_worker(workers["dynamic_connected_busy"])
        wr.mark_running()

        workers["dynamic_connected_now_idle"] = self.playground.create_worker(
            fqdn="now_idle.connected.dynamic.lan", worker_pool=worker_pool
        )
        workers["dynamic_connected_now_idle"].connected_at = timezone.now()
        workers["dynamic_connected_now_idle"].save()
        wr = self.playground.create_worker_task(mark_pending=True)
        wr.assign_worker(workers["dynamic_connected_now_idle"])
        wr.mark_completed(WorkRequest.Results.SUCCESS)

        for worker, included in (
            ("static_disconnected", True),
            ("static_connected_idle", True),
            ("dynamic_terminated", False),
            ("dynamic_disconnected", True),
            ("dynamic_connected_idle", True),
            ("dynamic_connected_busy", True),
            ("dynamic_connected_now_idle", True),
        ):
            with self.subTest(worker=worker):
                self.assertEqual(
                    Worker.objects.active()
                    .filter(pk=workers[worker].pk)
                    .exists(),
                    included,
                )

    def test_waiting_for_work_request(self) -> None:
        """Test WorkerManager.waiting_for_work_request() return a Worker."""
        worker = Worker.objects.create_with_fqdn(
            "worker-a",
            token=self.playground.create_worker_token(set_worker=False),
        )

        # WorkerManagement.waiting_for_work_request: returns no workers because
        # the worker is not connected
        self.assertQuerySetEqual(Worker.objects.waiting_for_work_request(), [])

        worker.mark_connected()

        # The worker is not fully connected until it sends dynamic metadata
        self.assertQuerySetEqual(Worker.objects.waiting_for_work_request(), [])

        worker.set_dynamic_metadata({})

        # Now the Worker is ready to have a task assigned
        self.assertQuerySetEqual(
            Worker.objects.waiting_for_work_request(), [worker]
        )

        # A task is assigned to the worker
        work_request = self.playground.create_worker_task(
            worker=worker, mark_pending=True
        )

        # The worker is not ready (it is busy with a task assigned)
        self.assertQuerySetEqual(Worker.objects.waiting_for_work_request(), [])

        # The task finished
        work_request.status = WorkRequest.Statuses.COMPLETED
        work_request.save()

        # The worker is ready: the WorkRequest that had assigned finished
        self.assertQuerySetEqual(
            Worker.objects.waiting_for_work_request(), [worker]
        )

    def test_waiting_for_work_request_celery(self) -> None:
        """
        Test WorkerManager.waiting_for_work_request() with Celery workers.

        Celery workers have no tokens, so they require special handling.
        """
        worker = Worker.objects.get_or_create_celery()
        worker.concurrency = 3
        worker.save()

        # WorkerManagement.waiting_for_work_request: returns no workers because
        # the worker is not connected
        self.assertQuerySetEqual(Worker.objects.waiting_for_work_request(), [])

        worker.mark_connected()

        # Now the Worker is ready to have a task assigned
        self.assertQuerySetEqual(
            Worker.objects.waiting_for_work_request(), [worker]
        )

        # Assign two tasks to the worker; it can still accept more work
        for _ in range(2):
            work_request = self.playground.create_worker_task(
                worker=worker, mark_pending=True
            )
            self.assertQuerySetEqual(
                Worker.objects.waiting_for_work_request(), [worker]
            )

        # Assign a third task to the worker; it is now as busy as it allows
        work_request = self.playground.create_worker_task(
            worker=worker, mark_pending=True
        )
        self.assertQuerySetEqual(Worker.objects.waiting_for_work_request(), [])

        # The third task finished
        work_request.status = WorkRequest.Statuses.COMPLETED
        work_request.save()

        # Although it is still running some tasks, the worker can accept
        # more work again
        self.assertQuerySetEqual(
            Worker.objects.waiting_for_work_request(), [worker]
        )

    def test_waiting_for_work_request_no_return_disabled_workers(self) -> None:
        """Test WorkerManager.waiting_for_work_request() no return disabled."""
        worker_enabled = Worker.objects.create_with_fqdn(
            "worker-enabled",
            token=self.playground.create_worker_token(set_worker=False),
        )
        worker_enabled.mark_connected()
        worker_enabled.set_dynamic_metadata({})

        worker_disabled = Worker.objects.create_with_fqdn(
            "worker-disabled",
            token=self.playground.create_worker_token(
                enabled=False, set_worker=False
            ),
        )
        worker_disabled.mark_connected()
        worker_disabled.set_dynamic_metadata({})

        self.assertQuerySetEqual(
            Worker.objects.waiting_for_work_request(), [worker_enabled]
        )

    def test_waiting_for_work_request_static_first(self) -> None:
        """Ensure waiting_for_work_request() returns static workers first."""
        worker_static = Worker.objects.create_with_fqdn(
            "worker-static",
            token=self.playground.create_worker_token(set_worker=False),
        )

        worker_cloud = Worker.objects.create_with_fqdn(
            "worker-cloud",
            token=self.playground.create_worker_token(set_worker=False),
        )
        worker_cloud.worker_pool = WorkerPool.objects.create(
            name="test",
            provider_account=(
                self.playground.create_cloud_provider_account_asset()
            ),
            registered_at=timezone.now(),
        )
        worker_cloud.save()

        worker_cloud.mark_connected()
        worker_cloud.set_dynamic_metadata({})
        worker_static.mark_connected()
        worker_static.set_dynamic_metadata({})

        self.assertQuerySetEqual(
            Worker.objects.waiting_for_work_request(),
            [worker_static, worker_cloud],
        )

    def test_create_with_fqdn_new_fqdn(self) -> None:
        """WorkerManager.create_with_fqdn() return a worker."""
        token = self.playground.create_worker_token(set_worker=False)
        worker = Worker.objects.create_with_fqdn(
            'a-new-and-unique-name', token=token
        )

        self.assertEqual(worker.name, 'a-new-and-unique-name')
        self.assertEqual(worker.token, token)
        self.assertIsNotNone(worker.pk)
        token.refresh_from_db()
        self.assertEqual(token._worker_id, worker.id)

    def test_create_with_fqdn_duplicate_fqdn(self) -> None:
        """
        WorkerManager.create_with_fqdn() return a worker.

        The name ends with -2 because 'connected-worker' is already used.
        """
        Worker.objects.create_with_fqdn(
            'connected-worker',
            token=self.playground.create_worker_token(set_worker=False),
        )

        token = self.playground.create_worker_token(set_worker=False)
        worker = Worker.objects.create_with_fqdn(
            'connected-worker', token=token
        )

        self.assertEqual(worker.name, 'connected-worker-2')
        self.assertEqual(worker.token, token)
        self.assertEqual(worker.worker_type, WorkerType.EXTERNAL)
        self.assertIsNotNone(worker.pk)
        token.refresh_from_db()
        self.assertEqual(token._worker_id, worker.id)

    def test_create_with_fqdn_external_no_token(self) -> None:
        """
        Trying to create an external worker with no token is an error.

        This test is here to ensure that we don't enter an infinite loop
        trying to find a unique name.
        """
        with self.assertRaisesRegex(
            IntegrityError, "db_worker_celery_or_token"
        ):
            Worker.objects.create_with_fqdn(  # type: ignore[call-overload]
                "no-token"
            )

    def test_create_pool_members(self) -> None:
        worker_pool = self.playground.create_worker_pool(name="pool")
        # 1 existing worker in the middle of our scheme
        Worker.objects.create_with_fqdn(
            "pool-002",
            token=self.playground.create_worker_token(set_worker=False),
        )

        before = timezone.now()
        Worker.objects.create_pool_members(worker_pool, 3)
        after = timezone.now()

        workers = Worker.objects.filter(
            worker_pool=worker_pool,
            activation_token__isnull=False,
            activation_token__expire_at__gte=(
                before + dt.timedelta(minutes=15)
            ),
            activation_token__expire_at__lte=(after + dt.timedelta(minutes=15)),
            activation_token__enabled=True,
            token__isnull=True,
        )
        self.assertQuerySetEqual(
            workers.order_by("name").values_list("name", flat=True),
            ["pool-001", "pool-003", "pool-004"],
        )
        self.assertQuerySetEqual(
            Token.objects.filter(activating_worker__in=workers).values_list(
                "_activating_worker_id", flat=True
            ),
            {worker.id for worker in workers},
            ordered=False,
        )

    def test_get_or_create_scheduler(self) -> None:
        """WorkerManager.get_or_create_scheduler returns a scheduler worker."""
        worker = Worker.objects.get_or_create_scheduler()

        self.assertEqual(worker.name, "scheduler")
        self.assertIsNone(worker.token)
        self.assertEqual(worker.worker_type, WorkerType.SCHEDULER)
        self.assertIsNotNone(worker.pk)

        self.assertEqual(Worker.objects.get_or_create_scheduler(), worker)

    def test_get_or_create_celery(self) -> None:
        """WorkerManager.get_or_create_celery returns a Celery worker."""
        worker = Worker.objects.get_or_create_celery()

        self.assertEqual(worker.name, "celery")
        self.assertIsNone(worker.token)
        self.assertEqual(worker.worker_type, WorkerType.CELERY)
        self.assertIsNotNone(worker.pk)

        self.assertEqual(Worker.objects.get_or_create_celery(), worker)

    def test_slugify_with_suffix_counter_1(self) -> None:
        """WorkerManager._generate_unique_name does not append '-1'."""
        self.assertEqual(
            WorkerManager._generate_unique_name('worker.lan', 1), 'worker-lan'
        )

    def test_slugify_with_suffix_counter_3(self) -> None:
        """WorkerManager._generate_unique_name appends '-3'."""
        self.assertEqual(
            WorkerManager._generate_unique_name('worker.lan', 3), 'worker-lan-3'
        )

    def test_get_worker_by_token_or_none_return_none(self) -> None:
        """WorkerManager.get_worker_by_token_or_none() return None."""
        self.assertIsNone(
            Worker.objects.get_worker_by_token_key_or_none('non-existing-key')
        )

    def test_get_worker_by_token_or_none_return_worker(self) -> None:
        """WorkerManager.get_worker_by_token_or_none() return the Worker."""
        token = self.playground.create_worker_token(set_worker=False)

        worker = Worker.objects.create_with_fqdn('worker-a', token=token)

        self.assertEqual(
            Worker.objects.get_worker_by_token_key_or_none(token.key), worker
        )

    def test_get_worker_or_none_return_worker(self) -> None:
        """WorkerManager.get_worker_or_none() return the Worker."""
        token = self.playground.create_worker_token(set_worker=False)

        worker = Worker.objects.create_with_fqdn('worker-a', token=token)

        self.assertEqual(Worker.objects.get_worker_or_none('worker-a'), worker)

    def test_get_worker_or_none_return_none(self) -> None:
        """WorkerManager.get_worker_or_none() return None."""
        self.assertIsNone(Worker.objects.get_worker_or_none('does-not-exist'))

    def test_with_idle_time_no_work_requests_not_connected(self) -> None:
        """Test idle_time with no work requests assigned and never connected."""
        worker = Worker.objects.create_with_fqdn(
            "worker-a",
            token=self.playground.create_worker_token(set_worker=False),
        )
        worker.registered_at = timezone.now() - dt.timedelta(days=1)
        worker.save()

        idle_time = (
            Worker.objects.with_idle_time()
            .values_list("idle_time", flat=True)
            .get(pk=worker.pk)
        )
        self.assertGreaterEqual(idle_time, 86400)
        self.assertLessEqual(idle_time, 86400 + 3600)

    def test_with_idle_time_no_work_requests_connected(self) -> None:
        """Compute idle_time with no work requests assigned and connected."""
        worker = Worker.objects.create_with_fqdn(
            "worker-a",
            token=self.playground.create_worker_token(set_worker=False),
        )
        worker.registered_at = timezone.now() - dt.timedelta(days=2)
        worker.instance_created_at = timezone.now() - dt.timedelta(days=1)
        worker.save()

        idle_time = (
            Worker.objects.with_idle_time()
            .values_list("idle_time", flat=True)
            .get(pk=worker.pk)
        )
        self.assertGreaterEqual(idle_time, 86400)
        self.assertLessEqual(idle_time, 86400 + 3600)

    def test_with_idle_time_running_task(self) -> None:
        """Compute idle_time with a running task."""
        worker = Worker.objects.create_with_fqdn(
            "worker-a",
            token=self.playground.create_worker_token(set_worker=False),
        )
        worker.registered_at = timezone.now() - dt.timedelta(days=2)
        worker.instance_created_at = timezone.now() - dt.timedelta(days=1)
        worker.save()
        self.playground.create_worker_task(worker=worker, mark_running=True)
        worker.refresh_from_db()
        self.assertTrue(worker.is_busy())

        idle_time = (
            Worker.objects.with_idle_time()
            .values_list("idle_time", flat=True)
            .get(pk=worker.pk)
        )
        self.assertEqual(idle_time, 0)

    def test_with_idle_time_pending_task(self) -> None:
        """Compute idle_time with a pending task."""
        worker = Worker.objects.create_with_fqdn(
            "worker-a",
            token=self.playground.create_worker_token(set_worker=False),
        )
        worker.registered_at = timezone.now() - dt.timedelta(days=2)
        worker.instance_created_at = timezone.now() - dt.timedelta(days=1)
        worker.save()
        self.playground.create_worker_task(worker=worker, mark_pending=True)
        worker.refresh_from_db()
        self.assertTrue(worker.is_busy())

        idle_time = (
            Worker.objects.with_idle_time()
            .values_list("idle_time", flat=True)
            .get(pk=worker.pk)
        )
        self.assertEqual(idle_time, 0)

    def test_with_idle_time_completed_task(self) -> None:
        """Compute idle_time with a completed task."""
        worker = Worker.objects.create_with_fqdn(
            "worker-a",
            token=self.playground.create_worker_token(set_worker=False),
        )
        worker.registered_at = timezone.now() - dt.timedelta(days=2)
        worker.instance_created_at = timezone.now() - dt.timedelta(days=1)
        worker.save()
        for hours_ago in (10, 6):
            work_request = self.playground.create_worker_task(
                worker=worker, result=WorkRequest.Results.SUCCESS
            )
            work_request.completed_at = timezone.now() - dt.timedelta(
                hours=hours_ago
            )
            work_request.save()

        idle_time = (
            Worker.objects.with_idle_time()
            .values_list("idle_time", flat=True)
            .get(pk=worker.pk)
        )
        self.assertGreaterEqual(idle_time, 6 * 3600)
        self.assertLessEqual(idle_time, 7 * 3600)

    def test_with_idle_time_old_completed_task(self) -> None:
        """Compute idle_time with a completed task from a previous instance."""
        worker = Worker.objects.create_with_fqdn(
            "worker-a",
            token=self.playground.create_worker_token(set_worker=False),
        )
        worker.registered_at = timezone.now() - dt.timedelta(days=7)
        worker.instance_created_at = timezone.now() - dt.timedelta(seconds=5)
        worker.save()
        for hours_ago in (10, 6):
            work_request = self.playground.create_worker_task(
                worker=worker, result=WorkRequest.Results.SUCCESS
            )
            work_request.completed_at = timezone.now() - dt.timedelta(
                hours=hours_ago
            )
            work_request.save()

        idle_time = (
            Worker.objects.with_idle_time()
            .values_list("idle_time", flat=True)
            .get(pk=worker.pk)
        )
        self.assertLessEqual(idle_time, 60)

    def test_mark_all_disconnected(self) -> None:
        """WorkerManager.mark_all_disconnected() affects non-Celery workers."""
        worker_connected_external = self.playground.create_worker(
            fqdn="connected-external"
        )
        worker_connected_external.mark_connected()
        worker_connected_signing = self.playground.create_worker(
            worker_type=WorkerType.SIGNING, fqdn="connected-signing"
        )
        worker_connected_signing.mark_connected()
        worker_connected_celery = Worker.objects.get_or_create_celery()
        worker_connected_celery.mark_connected()
        worker_disconnected_external = self.playground.create_worker(
            fqdn="disconnected-external"
        )
        worker_disconnected_signing = self.playground.create_worker(
            worker_type=WorkerType.SIGNING, fqdn="disconnected-signing"
        )

        Worker.objects.mark_all_disconnected()

        worker_connected_celery.refresh_from_db()
        self.assertTrue(worker_connected_celery.connected())
        for disconnected in (
            worker_connected_external,
            worker_connected_signing,
            worker_disconnected_external,
            worker_disconnected_signing,
        ):
            disconnected.refresh_from_db()
            self.assertFalse(disconnected.connected())


class WorkerTests(TestCase):
    """Tests for the Worker model."""

    worker: ClassVar[Worker]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up the Worker for the tests."""
        super().setUpTestData()
        cls.worker = Worker.objects.create_with_fqdn(
            "computer.lan",
            token=cls.playground.create_worker_token(set_worker=False),
        )
        cls.worker.static_metadata = {"os": "debian"}
        cls.worker.set_dynamic_metadata({"cpu_cores": "4"})
        cls.worker.save()

    def test_mark_connected(self) -> None:
        """Test mark_connected method."""
        time_before = timezone.now()
        self.assertIsNone(self.worker.connected_at)

        self.worker.mark_connected()
        assert self.worker.connected_at is not None

        self.assertGreaterEqual(self.worker.connected_at, time_before)
        self.assertLessEqual(self.worker.connected_at, timezone.now())

    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    def test_mark_connected_runs_scheduler(self) -> None:
        """mark_connected asks the scheduler to run."""
        self.assertIsNone(self.worker.connected_at)

        with (
            mock.patch("debusine.server.scheduler.schedule") as mock_schedule,
            self.captureOnCommitCallbacks(execute=True),
        ):
            self.worker.mark_connected()

        mock_schedule.assert_called_once()

    def test_mark_disconnected(self) -> None:
        """Test mark_disconnected method."""
        self.worker.mark_connected()
        self.assertTrue(self.worker.connected())

        self.worker.mark_disconnected()

        self.assertFalse(self.worker.connected())
        self.assertIsNone(self.worker.connected_at)

    def test_connected(self) -> None:
        """Test connected method."""
        self.assertFalse(self.worker.connected())

        self.worker.connected_at = timezone.now()

        self.assertTrue(self.worker.connected())

    def test_is_busy(self) -> None:
        """Test is_busy method."""
        worker = self.playground.create_worker(fqdn="test")
        self.assertFalse(worker.is_busy())

        work_request = self.playground.create_worker_task(
            worker=worker, mark_pending=True
        )

        self.assertEqual(work_request.status, WorkRequest.Statuses.PENDING)
        self.assertTrue(worker.is_busy())

        work_request.assign_worker(worker)

        self.assertTrue(worker.is_busy())

        work_request.mark_running()

        self.assertTrue(worker.is_busy())

        work_request.mark_aborted()

        self.assertFalse(worker.is_busy())

    def test_is_at_capacity(self) -> None:
        """The is_at_capacity method handles workers with concurrency > 1."""
        worker = Worker.objects.get_or_create_celery()
        worker.concurrency = 3
        worker.save()
        self.assertFalse(worker.is_at_capacity())
        for _ in range(3):
            self.assertFalse(worker.is_at_capacity())
            work_request = self.playground.create_worker_task(
                worker=worker, mark_pending=True
            )
        self.assertTrue(worker.is_at_capacity())
        work_request.mark_aborted()
        self.assertFalse(worker.is_at_capacity())

    def test_running_work_requests(self) -> None:
        """The running_work_requests return expected work requests."""
        worker_1 = self.playground.create_worker()
        worker_1.concurrency = 3
        worker_1.save()

        worker_2 = self.playground.create_worker()

        work_request_1 = self.playground.create_worker_task(
            worker=worker_1, mark_running=True
        )
        work_request_2 = self.playground.create_worker_task(
            worker=worker_1, mark_running=True
        )
        self.playground.create_worker_task(worker=worker_1, mark_running=False)
        self.playground.create_worker_task(worker=worker_2, mark_running=True)

        self.assertQuerySetEqual(
            worker_1.running_work_requests(), [work_request_1, work_request_2]
        )

    def test_metadata_no_conflict(self) -> None:
        """Test metadata method: return all the metadata."""
        self.assertEqual(
            self.worker.metadata(),
            {
                'cpu_cores': '4',
                'os': 'debian',
                'provided_tags': ['worker:type:external'],
            },
        )

    def test_metadata_with_conflict(self) -> None:
        """
        Test metadata method: return all the metadata.

        static_metadata has priority over dynamic_metadata
        """
        # Assert initial state
        self.assertEqual(self.worker.dynamic_metadata['cpu_cores'], '4')

        # Add new static_metadata key
        self.worker.static_metadata['cpu_cores'] = '8'

        self.assertEqual(
            self.worker.metadata(),
            {
                'cpu_cores': '8',
                'os': 'debian',
                'provided_tags': ['worker:type:external'],
            },
        )

    def test_metadata_merge_provided_tags(self) -> None:
        """Test metadata method: provided_tags are merged."""
        self.worker.dynamic_metadata['provided_tags'] = ['foo', 'bar']
        self.worker.static_metadata['provided_tags'] = ['bar', 'baz']

        metadata = self.worker.metadata()
        self.assertEqual(
            metadata['provided_tags'],
            ['bar', 'baz', 'foo', 'worker:type:external'],
        )

    def test_metadata_is_deep_copy(self) -> None:
        """Test metadata does a deep copy."""
        self.worker.dynamic_metadata['list'] = ['foo', 'bar']
        self.worker.metadata()['list'].append('baz')

        self.assertEqual(self.worker.dynamic_metadata['list'], ['foo', 'bar'])

    def test_set_dynamic_metadata(self) -> None:
        """Worker.set_dynamic_metadata sets the dynamic metadata."""
        self.worker.dynamic_metadata = {}
        self.worker.dynamic_metadata_updated_at = None
        self.worker.save()

        dynamic_metadata = {"cpu_cores": 4, "ram": "16"}
        self.worker.set_dynamic_metadata(dynamic_metadata)
        assert self.worker.dynamic_metadata_updated_at is not None

        self.worker.refresh_from_db()

        self.assertEqual(self.worker.dynamic_metadata, dynamic_metadata)
        self.assertLessEqual(
            self.worker.dynamic_metadata_updated_at, timezone.now()
        )

    def test_str(self) -> None:
        """Test WorkerTests.__str__."""
        self.assertEqual(
            self.worker.__str__(),
            f"Id: {self.worker.id} Name: {self.worker.name}",
        )

    def test_ui(self) -> None:
        from debusine.web.views.ui.workers import WorkerUI

        request = RequestFactory().get("/")
        ui = self.worker.ui(request)
        self.assertIsInstance(ui, WorkerUI)

    def test_get_absolute_url(self) -> None:
        """Test the get_absolute_url method."""
        self.assertEqual(
            self.worker.get_absolute_url(),
            f"/-/status/workers/{self.worker.name}/",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""utils module contain utilities used by different components of debusine."""

import hashlib
import os
import re
import shutil
import tempfile
from collections.abc import Callable, Generator, Mapping, Sequence
from contextlib import contextmanager
from enum import StrEnum
from enum import property as enum_property
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path, PurePath
from types import GenericAlias
from typing import (
    Annotated,
    Any,
    BinaryIO,
    IO,
    Literal,
    TYPE_CHECKING,
    TextIO,
    TypedDict,
    Union,
    overload,
)

import pydantic
from debian import deb822

# https://github.com/pydantic/pydantic/issues/12720
from pydantic._internal._generics import get_args, get_origin
from requests.structures import CaseInsensitiveDict

if TYPE_CHECKING:
    from _typeshed import OpenBinaryModeWriting, OpenTextModeWriting
    from django.http.request import HttpHeaders

CALCULATE_HASH_CHUNK_SIZE = 1 * 1024 * 1024


def calculate_hash(file_path: Path, hash_name: str) -> bytes:
    """Return hash (using algorithm hash_name) of file."""
    hasher = hashlib.new(hash_name)

    with open(file_path, "rb") as f:
        while chunk := f.read(CALCULATE_HASH_CHUNK_SIZE):
            hasher.update(chunk)

    return hasher.digest()


# taken from functional.py in python3-django 3:4.2.11-1 to avoid the dependency
class classproperty[Get]:
    """
    classproperty definition.

    Decorator that converts a method with a single cls argument into a property
    that can be accessed directly from the class.
    """

    def __init__(self, method: Callable[[Any], Get]) -> None:
        """Define __init__."""
        self.fget = method

    def __get__(
        self,
        instance: Any | None,  # noqa: U100
        cls: type[Any] | None = None,
    ) -> Get:
        """Define __get__."""
        return self.fget(cls)


def _error_message_invalid_header(header_name: str, header_value: str) -> str:
    return f'Invalid {header_name} header: "{header_value}"'


# TODO: RFC 7233 allows multiple byte-ranges separated by commas, but this
# currently raises an error for those.
def parse_range_header(
    headers: Union[CaseInsensitiveDict[str], "HttpHeaders"], file_size: int
) -> dict[str, int] | None:
    """Parse headers["Range"]. Return dictionary with information."""
    header_name = "Range"
    header_value = headers.get(header_name)

    if header_value is None:
        return None

    if m := re.match("^bytes=([0-9]*)-([0-9]*)$", header_value):
        start, end = m.groups()
        if start or end:
            if not start:
                # This is a suffix-byte-range-spec
                # (https://www.rfc-editor.org/rfc/rfc7233#section-2.1), and
                # requests the last ``end`` bytes of the representation (not
                # from the start of the representation to the byte with
                # offset ``end``, as one might expect).
                start = file_size - int(end)
                end = file_size - 1
            elif not end:
                # Request from ``start`` to the end of the file.
                end = file_size - 1
            return {"start": int(start), "end": int(end)}

    raise ValueError(_error_message_invalid_header(header_name, header_value))


class ParsedContentRange(TypedDict):
    """A parsed Content-Range header."""

    start: int | Literal["*"]
    end: int | None
    size: int | Literal["*"]


def parse_content_range_header(
    headers: Mapping[str, str],
) -> ParsedContentRange | None:
    """Parse headers["Content-Range"]. Return dictionary with information."""
    header_name = "Content-Range"
    header_value = headers.get("Content-Range")

    if header_value is None:
        return None

    if m := re.match(r"bytes ([0-9]+)-([0-9]+)/([0-9]+|\*)", header_value):
        return {
            "start": int(m.group(1)),
            "end": int(m.group(2)),
            "size": "*" if m.group(3) == "*" else int(m.group(3)),
        }
    elif m := re.match(r"bytes \*/([0-9]+)", header_value):
        return {
            "start": "*",
            "end": None,
            "size": int(m.group(1)),
        }
    elif re.match(r"bytes \*/\*", header_value):
        return {
            "start": "*",
            "end": None,
            "size": "*",
        }
    raise ValueError(_error_message_invalid_header(header_name, header_value))


def is_single_path_component(value: str) -> bool:
    """Return whether ``value`` is a single path component."""
    return PurePath(value).name == value


def _check_deb822_security(path: Path, data: deb822.Deb822) -> None:
    """Check that a deb822 file does not reference files outside the CWD."""
    for field in ("Files", "Checksums-Sha1", "Checksums-Sha256"):
        for file in data.get(field, []):
            if not is_single_path_component(file["name"]):
                raise ValueError(
                    f"{path.name!r} has an entry in {field} with more than one"
                    f" path component: {file['name']!r}"
                )


def read_dsc(dsc_path: Path) -> deb822.Dsc:
    """
    Read a .dsc file and return the contents.

    :raises ValueError: if the dsc does not have at least "source" and
      "version".
    """
    with open(dsc_path) as dsc_file:
        dsc = deb822.Dsc(dsc_file)

        if "source" not in dsc:
            raise ValueError(f"{dsc_path.name!r} does not contain Source field")
        if "version" not in dsc:
            raise ValueError(
                f"{dsc_path.name!r} does not contain Version field"
            )

        _check_deb822_security(dsc_path, dsc)

    return dsc


def read_changes(changes_path: Path) -> deb822.Changes:
    """Read a .changes file and return the contents."""
    with open(changes_path) as changes_file:
        changes = deb822.Changes(changes_file)

        _check_deb822_security(changes_path, changes)

    return changes


def find_files_suffixes(
    directory: Path, endswith: Sequence[str], *, include_symlinks: bool = False
) -> list[Path]:
    """
    Return files (sorted) ending with any of the endswith in directory.

    :param directory: directory where to search the files
    :param endswith: suffix to return the files from the directory
    :param include_symlinks: if False (default): does not return symbolic links,
      if True return symbolic links

    Find only regular files (no symbolic links, directories, etc.).
    """
    found_files: list[Path] = []

    for file in directory.iterdir():
        if not include_symlinks:
            if file.is_symlink():
                continue

        if not file.is_file():
            continue

        if file.name.endswith(tuple(endswith)):
            found_files.append(file)

    return sorted(found_files)


def find_file_suffixes(directory: Path, endswith: Sequence[str]) -> Path | None:
    """
    Find and return file ending with any of the endswith in directory.

    Finds regular files (no symbolic links, directories, etc.).

    Raise RuntimeError if more than one file could be returned.
    """
    found_files = find_files_suffixes(directory, endswith)

    number_of_files = len(found_files)

    if number_of_files == 1:
        return found_files[0]
    elif number_of_files == 0:
        return None
    else:
        found_files.sort()
        found_paths = [str(f) for f in found_files]
        raise RuntimeError(f"More than one {endswith} file: {found_paths}")


def is_command_available(cmd: str) -> bool:
    """
    Check whether cmd is available on $PATH.

    :param cmd: command name to check, passed to shutil.which.  (This may be
      a full path name, in which case shutil.which simply checks whether an
      executable exists at that path.)
    """
    return shutil.which(cmd) is not None


# The return type is really typing._GenericAlias, but that's private.
def _get_specialization(cls: type, base: type) -> Any:
    """
    Inspect a class for its specialization of a generic base class.

    A class may specialize a generic base class in various ways: for
    example, it might supply specific values for some of the relevant type
    variables, or it might itself be generic and rely on being specialized
    by its own subclasses, or both.  To help us introspect such classes,
    this returns the specialization of `base` by `cls` in the form of a
    :py:class:`typing._GenericAlias`.

    :raises AssertionError: if `cls` specializes `base` in different ways by
      means of multiple inheritance.
    """
    specializations: set[GenericAlias] = set()
    if hasattr(cls, "__pydantic_generic_metadata__"):
        orig_bases = [get_origin(cls)[get_args(cls)]]
    else:
        orig_bases = getattr(cls, "__orig_bases__", [])
    for cls_base in orig_bases:
        origin = get_origin(cls_base)
        if origin == base:
            specializations.add(cls_base)
        elif origin is not None and issubclass(origin, base):
            specializations.add(
                _get_specialization(origin, base)[get_args(cls_base)]
            )
    if len(specializations) != 1:
        raise AssertionError(
            f"{cls.__qualname__} must specialize {base.__qualname__} with "
            f"exactly one consistent list of type arguments "
            f"(got {specializations})"
        )
    [specialization] = specializations
    return specialization


def extract_generic_type_arguments(
    cls: type, expected_origin: type
) -> tuple[type, ...]:
    """
    Extract type arguments from a generic class.

    This is expected to be called from __init_subclass__ in a generic class
    (i.e. one that has Generic[...] as a base class), and allows extracting
    the specializing type arguments so that they can be used as factories.
    """
    args = get_args(_get_specialization(cls, expected_origin))
    assert isinstance(args, tuple)
    return args


class DjangoChoicesEnum(StrEnum):
    """Enable a StrEnum to be used for a Django choices field."""

    @enum_property
    def label(self) -> str:
        """Return this item's label."""
        return self

    @classproperty
    def choices(cls) -> Generator[tuple[str, str]]:
        """Return a Django model compatible set of choices."""
        for item in cls:
            # mypy can't quite follow the maze of enum properties here, but
            # this works.
            yield (item, item.label)  # type: ignore[attr-defined]


class NotSupportedError(Exception):
    """The requested function is not supported."""


@overload
@contextmanager
def atomic_writer(
    path: str | os.PathLike[str],
    mode: "OpenBinaryModeWriting" = "wb",
    chmod: int = ...,
) -> Generator[BinaryIO]: ...


@overload
@contextmanager
def atomic_writer(
    path: str | os.PathLike[str], mode: "OpenTextModeWriting", chmod: int = ...
) -> Generator[TextIO]: ...


@contextmanager
def atomic_writer(
    path: str | os.PathLike[str],
    mode: Union["OpenTextModeWriting", "OpenBinaryModeWriting"] = "wb",
    chmod: int = 0o600,
) -> Generator[IO[Any]]:
    """
    Atomically write to a file.

    The yielded file object is a temporary file in the same directory as the
    target path.  On exit from the context manager, it is renamed into
    place.
    """
    path = Path(path)
    with tempfile.NamedTemporaryFile(
        mode, dir=path.parent, prefix=path.name, suffix=".new", delete=False
    ) as temp:
        try:
            yield temp
            temp.flush()
            os.fdatasync(temp)
            os.fchmod(temp.fileno(), chmod)
            os.replace(temp.name, path)
        except Exception:
            os.unlink(temp.name)
            raise


def debusine_version() -> str | None:
    """Return Debusine's version, if available."""
    try:
        return version("debusine")
    except PackageNotFoundError:
        return None


def get_user_agent() -> str:
    """Return the User-Agent string to be used for HTTP requests."""
    version = debusine_version()
    if version:
        return f"debusine/{version}"
    return "debusine"


Base64BytesAsStr = Annotated[
    pydantic.Base64Bytes,
    pydantic.BeforeValidator(
        lambda x: x.encode(), json_schema_input_type="str"
    ),
    pydantic.WrapSerializer(
        lambda value, handler, info: handler(value, info).decode(),
        return_type="str",
    ),
]
